/*
 * Copyright notice from original mutt:
 * Copyright (C) ????-2004 Justin Hibbits <jrh29@po.cwru.edu>
 * Copyright (C) 2004 Thomer M. Gil <mutt@thomer.com>
 *
 * Parts were written/modified by:
 * Rocco Rutte <pdmef@cs.tu-berlin.de>
 * Nico Golde <nico@ngolde.de>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-ui/lib-ui.h>
#include <lib-ui/menu.h>

#include <libgen.h>

#include "mutt.h"
#include "charset.h"
#include "buffy.h"
#include "keymap.h"

static int CurBuffy = 0;

/* computes first entry to be shown */
static int calc_boundaries(void)
{
    if (CurBuffy < 0 || CurBuffy >= Incoming.len)
        CurBuffy = 0;

    if (option(OPTSIDEBARNEWMAILONLY)) {
        int n = 0;

        for (int i = 0; i < CurBuffy; i++) {
            n += Incoming.arr[i]->new > 0;
        }

        n %= (LINES - 3);
        if (!n)
            return CurBuffy;

        for (int i = CurBuffy - 1; i >= 0; i--) {
            if (Incoming.arr[i]->new > 0 && --n <= 0) {
                return i;
            }
        }

        return 0;
    } else {
        return CurBuffy - (CurBuffy % (LINES - 3));
    }
}

static void shortened_hierarchy(char *dst, char *hbox, int maxlen)
{
    int dots = 0, last_dot = 0, len = m_strlen(hbox);

    if (m_strisempty(SidebarBoundary)) {
        memcpy(dst, hbox, MIN(len, maxlen));
        return;
    }

    for (int i = 0; i < len; ++i) {
        if (strchr(SidebarBoundary, hbox[i])) {
            ++dots;
            last_dot = i;
        }
    }

    if (dots == 0) {
        memcpy(dst, hbox, MIN(len, maxlen));
        return;
    }

    dst[0] = hbox[0];
    for (int i = 1, j = 1; i < len && j < maxlen - 1; ++i) {
        if (!strchr(SidebarBoundary, hbox[i]))
            continue;

        if (len - i <= maxlen - j) {
            memcpy(dst, hbox + i, len - i);
            return;
        }

        if (i == last_dot) {
            memcpy(dst, hbox + i, maxlen - j);
            return;
        }

        dst[j++] = hbox[i];
        dst[j++] = hbox[i + 1];
    }
}

static const char *
sidebar_number_format(char* dest, ssize_t destlen,
                      char op, const char* src, const char* fmt,
                      const char* ifstr, const char* elstr,
                      anytype data, format_flag flags)
{
  char tmp[STRING];
  BUFFY* b = Incoming.arr[data.i];
  int opt = flags & M_FORMAT_OPTIONAL;
  int c = Context && !m_strcmp(Context->path, b->path);

  switch (op) {
    /* deleted */
    case 'd':
      if (!opt) {
        snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
        snprintf (dest, destlen, tmp, c ? Context->deleted : 0);
      } else if ((c && Context->deleted == 0) || !c)
        opt = 0;
      break;
    /* flagged */
    case 'F':
    case 'f':                   /* for compatibility */
      if (!opt) {
        snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
        snprintf (dest, destlen, tmp, c ? Context->flagged : b->msg_flagged);
      } else if ((c && Context->flagged == 0) || (!c && b->msg_flagged == 0))
        opt = 0;
      break;
    /* total */
    case 'c':                   /* for compatibility */
    case 'm':
      if (!opt) {
        snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
        snprintf (dest, destlen, tmp, c ? Context->msgcount : b->msgcount);
      } else if ((c && Context->msgcount == 0) || (!c && b->msgcount == 0))
        opt = 0;
      break;
    /* total shown, i.e. not hidden by limit */
    case 'M':
      if (!opt) {
        snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
        snprintf (dest, destlen, tmp, c ? Context->vcount : 0);
      } else if ((c && Context->vcount == 0) || !c)
        opt = 0;
      break;
    /* new */
    case 'n':
      if (!opt) {
        snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
        snprintf (dest, destlen, tmp, c ? Context->new : b->new);
      } else if ((c && Context->new == 0) || (!c && b->new == 0))
        opt = 0;
      break;
    /* unread */
    case 'u':
      if (!opt) {
        snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
        snprintf (dest, destlen, tmp, c ? Context->unread : b->msg_unread);
      } else if ((c && Context->unread == 0) || (!c && b->msg_unread == 0))
        opt = 0;
      break;
    /* tagged */
    case 't':
      if (!opt) {
        snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
        snprintf (dest, destlen, tmp, c ? Context->tagged : 0);
      } else if ((c && Context->tagged == 0) || !c)
        opt = 0;
      break;
  }

  if (flags & M_FORMAT_OPTIONAL)
    m_strformat(dest, destlen, 0, opt ? ifstr : elstr,
                sidebar_number_format, data, flags);
  return src;
}

/* print single item
 * returns:
 *      0       item was not printed ('cause of $sidebar_newmail_only)
 *      1       item was printed
 */
static int make_sidebar_entry(WINDOW *sw, char *sbox, int idx, ssize_t len)
{
    int lencnt = 0;
    char no[STRING], entry[STRING];
    int l_m = m_strlen(Maildir);

    if (option(OPTSIDEBARNEWMAILONLY) && sbox && Context && Context->path
    &&  m_strcmp(Context->path, sbox) && Incoming.arr[idx]->new == 0
    &&  idx != CurBuffy)
        return 0;

    m_strformat(no, sizeof(no), len, SidebarNumberFormat,
                sidebar_number_format, idx, 0);
    lencnt = m_strlen(no);

    if (l_m > 0 && m_strncmp(sbox, Maildir, l_m) == 0 && m_strlen(sbox) > l_m)
    {
        sbox += l_m;
        if (Maildir[strlen(Maildir) - 1] != '/') {
            sbox += 1;
        }
    } else {
        sbox = basename(sbox);
    }

    snprintf(entry, sizeof(entry), "%*s", (int)len, no);
    if (option(OPTSHORTENHIERARCHY) && m_strlen(sbox) > len - lencnt - 1) {
        shortened_hierarchy(entry, sbox, len - lencnt - 1);
    } else {
        memcpy(entry, sbox, MIN(len - lencnt - 1, m_strlen(sbox)));
    }
    waddnstr(sw, entry, len);

    return 1;
}

/* returns folder name of currently 
 * selected folder for <sidebar-open>
 */
const char *sidebar_get_current(void)
{
    if (0 <= CurBuffy && CurBuffy < Incoming.len)
        return Incoming.arr[CurBuffy]->path;
    return NULL;
}

/* internally sets item to buf */
void sidebar_set_current(const char* buf)
{
    int i = buffy_lookup(buf);
    if (i >= 0) {
        CurBuffy = i;
    }
}

/* fix counters for a context
 * FIXME since ctx must not be of our business, move it elsewhere
 */
void sidebar_set_buffystats(CONTEXT *curContext)
{
    int i = 0;
    BUFFY *tmp;

    if (!curContext || (i = buffy_lookup(curContext->path)) < 0)
        return;
    tmp = Incoming.arr[i];
    tmp->new         = curContext->new;
    tmp->msg_unread  = curContext->unread;
    tmp->msgcount    = curContext->msgcount;
    tmp->msg_flagged = curContext->flagged;
}

int sidebar_draw(void)
{
    static short prev_show_value = -1;
    int x, y, line;
    WINDOW *sw;

    /* initialize first time */
    if (prev_show_value != option(OPTMBOXPANE)) {
        if ((prev_show_value = option(OPTMBOXPANE))) {
            /* after toggle: force recounting of all mail */
            buffy_check(2);
        }
    }

    sw = ui_layout_sidebar_w();
    if (!sw)
        return 0;
    getmaxyx(sw, y, x);

    SETCOLOR(sw, MT_COLOR_SIDEBAR);
    mvwhline(sw, 0, 0, ACS_HLINE, x - 1);
    mvwaddch(sw, 0, x - 1, ACS_TTEE);

    line = 1;
    for (int i = calc_boundaries(); i < Incoming.len && line < y - 1; i++) {
        BUFFY *tmp = Incoming.arr[i];

        if (i == CurBuffy)
            SETCOLOR(sw, MT_COLOR_INDICATOR);
        else if (tmp->new > 0)
            SETCOLOR(sw, MT_COLOR_NEW);
        else if (tmp->msg_flagged > 0)
            SETCOLOR(sw, MT_COLOR_FLAGGED);
        else
            SETCOLOR(sw, MT_COLOR_NORMAL);

        if (make_sidebar_entry(sw, tmp->path, i, x - 1)) {
            SETCOLOR(sw, MT_COLOR_SIDEBAR);
            waddch(sw, ACS_VLINE);
            line++;
        }
    }

    while (line < y - 1) {
        SETCOLOR(sw, MT_COLOR_NORMAL);
        whline(sw, ' ', x - 1);
        SETCOLOR(sw, MT_COLOR_SIDEBAR);
        mvwaddch(sw, line++, x - 1, ACS_VLINE);
    }

    SETCOLOR(sw, MT_COLOR_SIDEBAR);
    mvwhline(sw, y - 1, 0, ACS_HLINE, x - 1);
    mvwaddch(sw, y - 1, x - 1, ACS_BTEE);
    return 0;
}

void sidebar_scroll(int op)
{
    if (!Incoming.len)
        return;

    switch (op) {
      case OP_SIDEBAR_NEXT:
        if (!option(OPTSIDEBARNEWMAILONLY)) {
            CurBuffy = (CurBuffy + 1) % Incoming.len;
            break;
        }
        /* FALLTHROUGH */
      case OP_SIDEBAR_NEXT_NEW:
        for (int i = 1; i < Incoming.len; i++) {
            if (Incoming.arr[(CurBuffy + i) % Incoming.len]->new > 0) {
                CurBuffy = (CurBuffy + i) % Incoming.len;
                break;
            }
        }
        break;

      case OP_SIDEBAR_PREV:
        if (!option(OPTSIDEBARNEWMAILONLY)) {
            if (!CurBuffy) {
                CurBuffy = Incoming.len;
            }
            CurBuffy--;
            break;
        }
        /* FALLTHROUGH */
      case OP_SIDEBAR_PREV_NEW:
        for (int i = Incoming.len - 1; i > 0; i--) {
            if (Incoming.arr[(CurBuffy + i) % Incoming.len]->new > 0) {
                CurBuffy = (CurBuffy + i) % Incoming.len;
                break;
            }
        }
        break;

      case OP_SIDEBAR_SCROLL_UP:
        CurBuffy -= LINES - 3;
        if (CurBuffy < 0)
            CurBuffy = 0;
        break;
      case OP_SIDEBAR_SCROLL_DOWN:
        CurBuffy += LINES - 3;
        if (CurBuffy >= Incoming.len)
            CurBuffy = Incoming.len - 1;
        break;
      default:
        return;
    }

    sidebar_draw();
}
